#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

# QUB Word Reports -- Feb 2004 Chris Nicolai -- cnicolai@buffalo.edu

# If the histograms don't show on your computer:
#   * run PythonWin ( Start -> Programs -> Python 2.1 -> PythonWin
#   * from the Tools menu, choose "COM MakePy Utility"
#   * double-click "Microsoft Excel 9.0 Object Library (1.3)"

import sys
import QUB
import QUBB
import Reports
import threading
import os.path
import string
import heapq
import traceback
import time

from math import *

from HistogramCommands import MakeHistogramSheet

import win32com.client
from pythoncom import CoInitialize, CoUninitialize



CoInitialize()


def setNbox(cell, val):
    cell.Value = val
    cell.Borders.LineStyle = 1 # xlContinuous

class WordReport:
    def __init__(self):
        self.word = win32com.client.Dispatch("Word.Application")
        self.word.Documents.Add()
        self.doc = self.word.Documents.Item( 1 )
        PSetup = self.doc.PageSetup
        PSetup.LeftMargin   = 20
        PSetup.RightMargin  = 20
        PSetup.TopMargin    = 20
        PSetup.BottomMargin = 20

        self.excel = win32com.client.Dispatch("Excel.Application")
        self.excel.DisplayAlerts = 0
        self.book = self.excel.Workbooks.Add(-4167) # XLWBatWorksheet
        self.scrap = self.book.Worksheets.Item(1)
        self.scrap.Cells.HorizontalAlignment = -4108 # 0xffffeff4 # xlHAlignCenter

        self.word.Visible = 1
        self.excel.Visible = 1

    def __del__(self):
        try:
            self.show()
            self.book.Close()
            self.excel.DisplayAlerts = 1
        finally:
            del self.scrap
            del self.book
            del self.excel
            del self.doc
            del self.word
        
    def show(self):
        self.word.Visible = 1

    def hide(self):
        self.word.Visible = 0

    def goToEnd(self):
        self.doc.Paragraphs.Add()
        npars = self.doc.Paragraphs.Count
        return self.doc.Range( self.doc.Paragraphs.Item(npars).Range.Start,
                               self.doc.Paragraphs.Item(npars).Range.End )
    
    def write(self, x='', align=0): # left-aligned (other choices: 1 center, 2 right, 3..8 various justifications)
        spot = self.goToEnd()
        spot.Text = str(x)
        spot.Paragraphs.Alignment = align
        return spot

    def paste(self, special=0, type=0):
        spot = self.goToEnd()

        if special:
            spot.PasteSpecial(0,0,0,0,type)
        else:
            spot.Paste()
            
        return spot

    def setColumns(self, ncol):
        self.goToEnd().InsertBreak(3) # wdSectionBreakContinuous
        self.doc.Sections.Item( self.doc.Sections.Count
                                ).PageSetup.TextColumns.SetCount(ncol);

    def pasteModelImage(self, modelNode=None, **cfg):
        # optional keywords override dialog box settings:
        #   [int]  Width, Height;
        #   [bool] DrawFrame, ShowTitles, Landscape, Bitmap
        try:
            currModel = QUB.ModelView.currModel
            if modelNode:
                storedModel = currModel.getAsNode()
                currModel.setAsNode( modelNode )
            currModel.copyImage( cfg )
            if modelNode:
                currModel.setAsNode( storedModel )

            if ('Enhanced' in cfg.keys()) and cfg['Enhanced']:
                self.paste(1, 0x9) # wdPasteEnhancedMetafile
            else:
                self.paste()

            mdl = self.doc.InlineShapes.Item( self.doc.InlineShapes.Count )
            #mdl.Width = width
            #mdl.Height = height
            mdl.Range.Paragraphs.Alignment = 1 # wdAlignParagraphCenter

            self.write('')
        
        except QUB.GraphicalInterrupt:
            raise
        else:
            pass
        
    def pasteDataImage(self, **cfg):
        # optional keywords override dialog box settings:
        #   [int]  Width, Height;
        #   [bool] DrawData, DrawIDLData, DrawBase,
        #          ScaleBars, ScaleBarsTop
        #          DrawFrame, ShowTitles, Landscape, Bitmap, HiRes
        QUB.DataView.CopyImage( cfg )

        if ('Enhanced' in cfg.keys()) and cfg['Enhanced']:
            self.paste(1, 0x9) # wdPasteEnhancedMetafile
        else:
            self.paste()

        try:
            dat = self.doc.InlineShapes.Item( self.doc.InlineShapes.Count )
            dat.Range.Paragraphs.Alignment = 1 # wdAlignParagraphCenter
        except QUB.GraphicalInterrupt:
            raise
        else:
            pass
        
        self.write()
        
    def writeTable(self, tbl, embed=0):
        sheet = self.book.Sheets.Add()
        cells = sheet.Cells
        maxcol = 0
        for row in tbl:
            if maxcol < len(row):
                maxcol = len(row)

        for i in range(0, len(tbl)):
            for j in range(0, len(tbl[i])):
                setNbox( cells(i+1, j+1), tbl[i][j] )
            for j in range(len(tbl[i]), maxcol):
                setNbox(cells(i+1, j+1), '')

        sel = sheet.Range( cells(1, 1), cells(len(tbl), maxcol) )
        sel.Copy()

        sel = self.goToEnd()
        if embed:
            sel.PasteSpecial(0,0,0,0,0)
        else:
            sel.Paste()

        sheet.Delete()
        return sel

    def writeHistogram(self, hist, w=300, h=200):
        sheet = self.book.Sheets.Add()
        chart = MakeHistogramSheet(hist, sheet, 1, int(w*1.4), int(h*1.4))

        chart.Width = w
        chart.Height = h

        chart.Chart.ChartArea.Copy();
        sel = self.goToEnd()
        sel.PasteSpecial(0, 0, 0)

        chart.Delete()
        sheet.Delete()

        return sel




def EnsureWordReport(context):
    try:
        win32com.client.gencache.EnsureModule('{00020813-0000-0000-C000-000000000046}',0,1,3)
    except:
        traceback.print_exc()

    try:
        win32com.client.gencache.EnsureModule('{00020905-0000-0000-C000-000000000046}',0,8,1)
    except:
        traceback.print_exc()

    try:
        return context.WordReport
    except AttributeError:
        context.WordReport = WordReport()
        return context.WordReport


# i'm an idjit -- i shoulda named 'em all "Properties" and now we all pay the price
def ResultsProperties(results):
    props = results.node.find('Properties')
    if not props.isNull:
        return props
    for name in ['AMP ', 'Stat ', 'MIL ', 'MPL ', 'AMP ', 'SKM ', 'MSearch ', '']:
        props = results.node.find(name+'Config')
        if not props.isNull:
            return props00


def ReportHeader(results, context):
    rep = EnsureWordReport(context)
    props = ResultsProperties(results)

    resultType = str(results.node['ResultType'].data)
    rep.write('QUB : %s' % resultType)
    rep.write( str(results.node['Date'].data) )

    if resultType == 'MIL':
        mode = str(props.find('Mode').data)
        segmode = str(props.find('use segments').data)
        if mode and segmode:
            rep.write(mode + ' segments ' + segmode)

    rep.write()

Reports.Add('Header', ReportHeader, ['MIL', 'MIP', 'Mac', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase'])


def ReportFileTable(results, context):
    rep = EnsureWordReport(context)

    resultType = str(results.node['ResultType'].data)
    if resultType == 'MIL':
        tbl = [ ['File Name', 'DeadTime', 'Channels', 'Events', 'LL', 'LL/event', 'in folder'] ]
        data = results.node.find('DataSet')
        while not data.isNull:
            dwellCount = 0
            LL = 0.0
            seg = data.find('Segment')
            while not seg.isNull:
                dwellCount += seg['DwellCount'].data[0]
                LL += seg['LL'].data[0]
                seg = seg.nextSameName()
            filename = str(data['FileName'].data)
            ccNode = data['ExpCond'].find('ChannelCount').data
            if len(data):
                channelCount = ccNode[0]
            else:
                channelCount = ''
                
            tbl.append( [os.path.basename(filename),
                         data['ExpCond']['DeadTime'].data[0],
                         channelCount,
                         dwellCount,
                         LL,
                         LL / dwellCount,
                         os.path.dirname(filename)] )
        
            data = data.nextSameName()
        rep.writeTable(tbl)
    else:
        tbl = [ ['File Name', 'Events', 'in folder'] ]
        data = results.node.find('DataSet')
        hasFilter = False
        if not data.find('ProcessData').isNull:
            tbl[0].append('Filter (Hz)')
            hasFilter = True
        while not data.isNull:
            dwellCount = 0
            seg = data.find('Segment')
            while not seg.isNull:
                try:
                    dwellCount += seg.find('DwellCount').data[0]
                except:
                    traceback.print_exc()
                seg = seg.nextSameName()
            filename = str(data['FileName'].data)

            entry = [os.path.basename(filename),
                     dwellCount,
                     os.path.dirname(filename)]
            
            if hasFilter:
                entry.append( str((data['ProcessData']['Filter'].data and
                                   data['ProcessData']['Filter'].data[0] and
                                   data['ProcessData']['FilterFreq'].data[0])
                                  or '') )

            tbl.append( entry )
            data = data.nextSameName()
        rep.writeTable(tbl)

Reports.Add('FileTable', ReportFileTable, ['MIL', 'MIP', 'Mac', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase'])


def ReportLLSummary(results, context):
    rep = EnsureWordReport(context)
    node = results.node
    if len(node.find('LL').data) == 0:
        return

    LL = node['LL'].data[0]
    LLperEvent = node['LL per event'].data[0]
    DwellCount = round(LL / LLperEvent)
    rep.write('LL: %f\tEvents: %d\tLL/event: %f\tIterations: %d' % (LL, DwellCount, LLperEvent, node['Iterations'].data[0]))
    rep.write('')

Reports.Add('LL Summary', ReportLLSummary, ['MIL', 'MPL', 'MIP', 'Mac', 'MSearch'])


def ReportMacConductance(results, context):
    rep = EnsureWordReport(context)
    node = results.node
    if len(node.find('ChannelCount std').data):
        rep.write('ChannelCount: %d +/- %f' % (node['ChannelCount'].data[0], node['ChannelCount std'].data[0]))
    if len(node.find('AmpOpen std').data):
        rep.write('AmpOpen: %f +/- %f' % (node['AmpOpen'].data[0], node['AmpOpen std'].data[0]))
    if len(node.find('VarOpen std').data):
        rep.write('VarOpen: %f +/- %f' % (node['VarOpen'].data[0], node['VarOpen std'].data[0]))

Reports.Add('Mac Conductance', ReportMacConductance, ['Mac'])


def ReportTimeConstants(results, context):
    rep = EnsureWordReport(context)
    node = results.node
    props = ResultsProperties(results)

    if str(props['use segments'].data) != 'together':
        return

    tc = node.find('DataSet').find('HistogramGroup').find('TimeConstants')
    if tc.isNull:
        return

    rep.write('Time Constants (Tau, Amp) of first file:')
    while not tc.isNull:
        line = "\tClass %d:" % (tc['Class'].data[0] + 1)
        tau = tc['Tau'].data
        amp = tc['Amp'].data
        for i in range(len(tau)):
            line = line + "\t(%.4f, %.4f)" % (tau[i], amp[i])
        rep.write(line)
        tc = tc.nextSameName()
    rep.write('')

Reports.Add('Time Constants', ReportTimeConstants, ['MIL', 'MPL'])


def ReportCriteria(results, context):
    rep = EnsureWordReport(context)
    for var in results.visibleVars:
        for color in results.usedColors:
            rep.write( str(var) + ',' + str(color) + ':'
                       + str(results[var].mean(color)) + ' ' + str(results[var].sd(color)) )

Reports.Add('Select: Stats', ReportCriteria, ['MIL', 'MPL', 'MIP', 'Mac', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase', 'MdlRpt'])


def ReportSearchParams(results, context):
    rep = EnsureWordReport(context)
    node = results.node
    props = ResultsProperties(results)

    dfp = props['DFP']

    tbl = [ ['LL Conv', 'Grad Conv', 'Max Step', 'Max Iterations', 'Max Restarts', 'Search Limit'],
            [dfp['ConvLL'].data[0], dfp['ConvGrad'].data[0],
             dfp['MaxStep'].data[0], dfp['MaxIterations'].data[0],
             dfp['MaxRestarts'].data[0], props['SearchLimit'].data[0]] ]
    rep.writeTable(tbl)

Reports.Add('Search Params', ReportSearchParams, ['MIL', 'MPL', 'MIP', 'Mac', 'MSearch'])


def writeModelImage(report, model):
    report.pasteModelImage( model,
                            Width=800, Height=200, Bitmap=0,
                            DrawFrame=1, ShowTitles=1, Landscape=1 )

def ReportInitModelImage(results, context):
    rep = EnsureWordReport(context)
    props = ResultsProperties(results)
    writeModelImage(rep, props['ModelFile'])

Reports.Add('Initial Model Image', ReportInitModelImage, ['MIL', 'MPL', 'MIP', 'MSearch', 'SKM', 'AMP', 'IdlBase', 'MdlRpt', 'Mac', 'MdlRpt'])


def WriteModelTable(report, model, includeSD=0):
    if includeSD:
        tbl = [ ['from', 'to', 'k0', 'k0std', 'k1', 'k1std', 'kEq'] ]
    else:
        tbl = [ ['from', 'to', 'k0', 'k1', 'kEq'] ]
    
    rate = model['Rates'].find('Rate')
    while not rate.isNull:
        states = rate['States'].data
        k0     = rate['k0'].data
        dk0    = rate['dk0'].data
        k1     = rate['k1'].data
        dk1    = rate['dk1'].data
        kEq    = k0[0] / k0[1]

        if includeSD:
            tbl.append( [states[0]+1, states[1]+1, '%.4g' % k0[0], '%.4g' % dk0[0], '%.4g' % k1[0], '%.4g' % dk1[0], '%.4g' % kEq] )
            tbl.append( [states[1]+1, states[0]+1, '%.4g' % k0[1], '%.4g' % dk0[1], '%.4g' % k1[1], '%.4g' % dk1[1], '%.4g' % (1.0/kEq)] )
        else:
            tbl.append( [states[0]+1, states[1]+1, '%.4g' % k0[0], '%.4g' % k1[0], '%.4g' % kEq] )
            tbl.append( [states[1]+1, states[0]+1, '%.4g' % k0[1], '%.4g' % k1[1], '%.4g' % kEq] )

        rate = rate.nextSameName()
    report.writeTable(tbl)

def WriteModelCalcTable(report, model, expcond, includeSD=0):
    if includeSD:
        cols = ['from', 'to', 'k0', 'k0std', 'k1', 'k1std', 'kEq', 'rate']
    else:
        cols = ['from', 'to', 'k0', 'k1', 'kEq', 'rate']
    stimVal = {}
    rate = model['Rates'].find('Rate')
    while not rate.isNull:
        for dir in (0, 1):
            for stim in ('P', 'Q'):
                if rate[stim].data[dir]:
                    name = str(rate[stim+'Names'][dir].data)
                    stimVal[name] = expcond[name].data[0]
        rate = rate.nextSameName()
    cols = cols + stimVal.keys()
    colOf = {}
    for i, col in enumerate(cols):
        colOf[col] = i
    colLetter = lambda i: chr( ord('A') + i )

    tbl = [ cols ]    
        
    def appendRate(frm, to, k0, dk0, k1, dk1, kEq, p, q):
        irow = len(tbl) + 1
        row = [''] * len(cols)
        row[colOf['from']] = frm
        row[colOf['to']] = to
        row[colOf['k0']] = '%.4g' % k0
        row[colOf['k1']] = '%.4g' % k1
        row[colOf['kEq']] = '%.4g' % kEq
        if includeSD:
            row[colOf['k0std']] = '%.4g' % dk0
            row[colOf['k1std']] = '%.4g' % dk1
        if irow == 2:
            for stim in stimVal.keys():
                row[colOf[stim]] = stimVal[stim]
        else:
            for stim in stimVal.keys():
                row[colOf[stim]] = '=%s%d' % (colLetter(colOf[stim]), 2)
                
        if p and q:
            row[colOf['rate']] = '=%s%d*%s%d*exp(%s%d*%s%d)' % (colLetter(colOf[p]), 2, colLetter(colOf['k0']), irow, colLetter(colOf[q]), 2, colLetter(colOf['k1']), irow)
        elif p:
            row[colOf['rate']] = '=%s%d*%s%d*exp(%s%d)' % (colLetter(colOf[p]), 2, colLetter(colOf['k0']), irow, colLetter(colOf['k1']), irow)
        elif q:
            row[colOf['rate']] = '=%s%d*exp(%s%d*%s%d)' % (colLetter(colOf['k0']), irow, colLetter(colOf[q]), 2, colLetter(colOf['k1']), irow)
        else:
            row[colOf['rate']] = '=%s%d*exp(%s%d)' % (colLetter(colOf['k0']), irow, colLetter(colOf['k1']), irow)
        tbl.append(row)
    
    rate = model['Rates'].find('Rate')
    while not rate.isNull:
        states = rate['States'].data
        k0     = rate['k0'].data
        dk0    = rate['dk0'].data
        k1     = rate['k1'].data
        dk1    = rate['dk1'].data
        kEq    = k0[1] / k0[0]
        p      = [rate['P'].data[i] and str(rate['PNames'][i].data) for i in range(2)]
        q      = [rate['Q'].data[i] and str(rate['QNames'][i].data) for i in range(2)]

        appendRate(states[0]+1, states[1]+1, k0[0], dk0[0], k1[0], dk1[0], kEq, p[0], q[0])        
        appendRate(states[1]+1, states[0]+1, k0[1], dk0[1], k1[1], dk1[1], kEq, p[1], q[1])        

        rate = rate.nextSameName()

    report.writeTable(tbl, embed=True)

def ReportInitModelTable(results, context):
    rep = EnsureWordReport(context)
    props = ResultsProperties(results)
    rep.write('Initial model:')
    WriteModelTable(rep, props['ModelFile'], 0)

Reports.Add('Initial Model Table', ReportInitModelTable, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'IdlBase', 'MIP', 'Mac', 'MdlRpt'])

def ReportInitModelCalcTable(results, context):
    rep = EnsureWordReport(context)
    props = ResultsProperties(results)
    rep.write('Initial model:')
    WriteModelCalcTable(rep, props['ModelFile'], results.node['DataSet']['ExpCond'], 0)

Reports.Add('Initial Model Calc Table', ReportInitModelCalcTable, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'IdlBase', 'MIP', 'Mac', 'MdlRpt'])


def writeModelConstraints(report, model):
    cns = model['Constraints'].child
    if not cns.isNull:
        report.write('Constraints:')
        while not cns.isNull:
            cnsLine = '\t' + cns.name + ':'
            for state in cns.data:
                cnsLine = cnsLine + ' ' + str(state+1)
            report.write( cnsLine )

            cns = cns.sibling

def ReportInitModelCns(results, context):
    rep = EnsureWordReport(context)
    props = ResultsProperties(results)
    writeModelConstraints(rep, props['ModelFile'])

Reports.Add('Initial Model Constraints', ReportInitModelCns, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'IdlBase', 'MdlRpt', 'MIP', 'Mac'])


def ReportFinalModelImage(results, context):
    rep = EnsureWordReport(context)
    writeModelImage(rep, results.node['ModelFile'])

Reports.Add('Final Model Image', ReportFinalModelImage, ['MIL', 'MPL', 'MSearch', 'IdlBase', 'MIP', 'Mac'])    


def ReportFinalModelTable(results, context):
    rep = EnsureWordReport(context)
    rep.write('Final model:')
    WriteModelTable(rep, results.node['ModelFile'], 1)

Reports.Add('Final Model Table', ReportFinalModelTable, ['MIL', 'MPL', 'MSearch', 'IdlBase', 'MIP', 'Mac'])    

def ReportFinalModelCalcTable(results, context):
    rep = EnsureWordReport(context)
    rep.write('Final model:')
    WriteModelCalcTable(rep, results.node['ModelFile'], results.node['DataSet']['ExpCond'], 1)

Reports.Add('Final Model Calc Table', ReportFinalModelCalcTable, ['MIL', 'MPL', 'MSearch', 'IdlBase', 'MIP', 'Mac'])    


def ReportFinalModelCns(results, context):
    rep = EnsureWordReport(context)
    writeModelConstraints(rep, results.node['ModelFile'])

Reports.Add('Final Model Constraints', ReportFinalModelCns, ['MIL', 'MPL', 'MSearch', 'IdlBase', 'MIP', 'Mac'])


def ApplyToAllByName(root, name, func):
    if root.name == name:
        func(root)
    child = root.child
    while not child.isNull:
        ApplyToAllByName(child, name, func)
        child = child.sibling


def writeNextTcritUp(report, hist):
    parent = hist.parent
    while not parent.isNull:
        tcrit = parent.find('Tcrit').data
        if len(tcrit):
            report.write('Tcrit: ' + string.join([str(tc) for tc in tcrit], '\t'))
            return
        parent = parent.parent

def writeHistAndTC(report, hist, tc, width, height):
    report.writeHistogram( hist, width, height )
    tcLine = 'Time Constants (Tau, Amp):'
    tau = tc['Tau'].data
    amp = tc['Amp'].data
    for i in range(0, len(tau)):
        tcLine = tcLine + (' (%f, %f)' % (tau[i], amp[i]))
    report.write( tcLine )
    if hist == hist.parent.find('Histogram'): # first child -- closed histogram
        writeNextTcritUp(report, hist)

def ReportHistograms(results, context):
    rep = EnsureWordReport(context)
    node = results.node

    def ReportHistGroup(group):
        rep.write(group['Title'].data)
        rep.setColumns(2)
        hist = group.find('Histogram')
        tc = group.find('TimeConstants')
        if not tc.isNull:
            while (not hist.isNull) and (not tc.isNull):
                writeHistAndTC(rep, hist, tc, 300, 200)
                hist = hist.nextSameName()
                tc = tc.nextSameName()
        else:
            while not hist.isNull:
                rep.writeHistogram(hist, 300, 200)
                hist = hist.nextSameName()
        rep.setColumns(1)

    ApplyToAllByName(node, 'HistogramGroup', ReportHistGroup)        

Reports.Add('Histograms', ReportHistograms, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase'])


def ReportLargeHistograms(results, context):
    rep = EnsureWordReport(context)
    node = results.node

    def ReportHistGroup(group):
        rep.write(group['Title'].data)
        hist = group.find('Histogram')
        tc = group.find('TimeConstants')
        if not tc.isNull:
            while (not hist.isNull) and (not tc.isNull):
                writeHistAndTC(rep, hist, tc, 540, 330)
                hist = hist.nextSameName()
                tc = tc.nextSameName()
        else:
            while not hist.isNull:
                rep.writeHistogram(hist, 540, 330)
                hist = hist.nextSameName()

    ApplyToAllByName(node, 'HistogramGroup', ReportHistGroup)        

Reports.Add('Large Histograms', ReportLargeHistograms, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase'])


#def ReportHistImages(results, context):
#    rep = EnsureWordReport(context)
#
#    for i in range(results.histogramCount):
#        results.copyHistogramImage(i)
#        rep.paste()

#Reports.Add('Histogram Images', ReportHistImages, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase'])


def ReportChartImages(results, context):
    rep = EnsureWordReport(context)

    for i in range(results.chartCount):
        results.copyChartImage(i)
        rep.paste()

Reports.Add('Select: Charts', ReportChartImages, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase', 'MdlRpt', 'MIP', 'Mac'])


def ReportSegTable(results, context):
    rep = EnsureWordReport(context)
    node = results.node

    haveIter = not node['DataSet']['Segment'].find('Iterations').isNull
    if haveIter:
        tbl = [ ['Segment', 'Bounds', 'Iterations', 'LogLikelihood', 'Events', 'LL per event', 'Gradient'] ]
    else:
        tbl = [ ['Segment', 'Bounds', 'LogLikelihood', 'Events', 'LL per event', 'Gradient'] ]

    segno = 1
    dset = node.find('DataSet')
    while not dset.isNull:
        seg = dset.find('Segment')
        while not seg.isNull:
            tbLine = [segno, ('%d - %d' % (seg.data[0], seg.data[1]))]
            if haveIter:
                tbLine.append( seg['Iterations'].data[0] )
            LL = seg['LL'].data[0]
            DwellCount = seg['DwellCount'].data[0]
            tbLine.append( LL )
            tbLine.append( DwellCount )
            tbLine.append( LL / DwellCount )
            grad = seg['Gradient'].data
            for g in grad:
                tbLine.append(g)
            tbl.append(tbLine)

            segno += 1
            seg = seg.nextSameName()
        dset = dset.nextSameName()
    rep.writeTable(tbl)

Reports.Add('Segment LL', ReportSegTable, ['MIL', 'MPL'])


def ReportSegRateTable(results, context):
    rep = EnsureWordReport(context)
    node = results.node
    props = ResultsProperties(results)

    tbl = []
    row = ['Segment', 'LL']
    activek1 = []
    
    rate   = props['ModelFile']['Rates'].find('Rate')
    while not rate.isNull:
        states = rate['States'].data
        k1     = rate['k1'].data
        there  = '%d_%d' % (states[0]+1, states[1]+1)
        back   = '%d_%d' % (states[1]+1, states[0]+1)
        
        row.append('k0_%s'  % there)
        row.append('k0s_%s' % there)
        row.append('k0_%s'  % back)
        row.append('k0s_%s' % back)

        if k1[0] != 0.0:
            activek1.append('k1_%s'  % there)
            activek1.append('k1s_%s' % there)
        if k1[1] != 0.0:
            activek1.append('k1_%s'  % back)
            activek1.append('k1s_%s' % back)
        
        rate = rate.nextSameName()

    tbl.append(row + activek1)

    segno = 1
    dset =node.find('DataSet')
    while not dset.isNull:
        seg = dset.find('Segment')
        while not seg.isNull:
            row = [segno, seg['LL'].data[0]]
            activek1 = []
            rate = seg['ModelFile']['Rates'].find('Rate')
            while not rate.isNull:
                k0     = rate['k0'].data
                dk0    = rate['dk0'].data
                k1     = rate['k1'].data
                dk1    = rate['dk1'].data

                row.append( k0[0] )
                row.append( dk0[0] )
                row.append( k0[1] )
                row.append( dk0[1] )

                if k1[0] != 0.0:
                    activek1.append( k1[0] )
                    activek1.append( dk1[0] )
                if k1[1] != 0.0:
                    activek1.append( k1[1] )
                    activek1.append( dk1[1] )

                rate = rate.nextSameName()
            tbl.append( row + activek1 )
            segno += 1
            seg = seg.nextSameName()
        dset = dset.nextSameName()
        
    rep.writeTable( tbl )

Reports.Add('Segment Rates', ReportSegRateTable, ['MIL', 'MPL'])


def makeOneSegTCLine(seg, segno, cls):
    line = [segno]
    tc = seg.find('HistogramGroup').find('TimeConstants')
    tcls = 0
    while tcls < cls:
        tc = tc.nextSameName()
        tcls += 1
    if not tc.isNull:
        tau = tc['Tau'].data
        amp = tc['Amp'].data
        for i in range(0, len(tau)):
            line.append( tau[i] )
            line.append( amp[i] )
    return line

def writeSegTCTable(report, results, cls):
    segno = 1
    dset = results.node.find('DataSet')
    props = ResultsProperties(results)

    if str(props['use segments'].data) == 'together':
        tbl = [ ['File', 'Tau', 'Amp', '...'] ]
        while not dset.isNull:
            tbl.append(makeOneSegTCLine(dset, segno, cls))
            segno += 1
            dset = dset.nextSameName()
    else:
        tbl = [ ['Segment', 'Tau', 'Amp', '...'] ]
        while not dset.isNull:
            seg = dset.find('Segment')
            while not seg.isNull:
                tbl.append(makeOneSegTCLine(seg, segno, cls))
                seg = seg.nextSameName()
                segno += 1
            dset = dset.nextSameName()

    report.write('Class %d:' % (cls+1))
    report.writeTable(tbl)

def getTCClassCount(container,prevmax):
    nclass = 0
    tc = container.find('TimeConstants')
    while not tc.isNull:
        nclass = nclass + 1
        tc = tc.nextSameName()
    if prevmax > nclass:
        nclass = prevmax
    return nclass

def getMaxClassCount(results):
    maxncls = 0
    dset = results.node.find('DataSet')
    while not dset.isNull:
        maxncls = getTCClassCount(dset.find('HistogramGroup'), maxncls)
            
        seg = dset.find('Segment')
        while not seg.isNull:
            maxncls = getTCClassCount(seg.find('HistogramGroup'), maxncls)
            seg = seg.nextSameName()
        dset = dset.nextSameName()
    return maxncls

def ReportSegTCTables(results, context):
    rep = EnsureWordReport(context)
    rep.write('Time Constants:')
    
    maxnclass = getMaxClassCount(results)
    for i in range(0, maxnclass):
        writeSegTCTable(rep, results, i)

Reports.Add('Segment Time Constants', ReportSegTCTables, ['MIL', 'MPL'])        

    
def ReportDataImage(results, context):
    rep = EnsureWordReport(context)
    rep.pasteDataImage()

Reports.Add('Data Image', ReportDataImage, ['MIL', 'MPL', 'MSearch', 'SKM', 'AMP', 'Stat', 'IV', 'ChopIdl', 'IdlBase', 'MIP', 'Mac'])

def BuildSegmentsTable(results, visVarsOnly=False):
    """Returns a list of rows, each a list of cells"""
    tbl = []
    if visVarsOnly:
        vars = results.visibleVars
    else:
        vars = results.keys()
    tbl.append( ['Segment', 'first', 'last', 'color'] + vars )
    for i in range(results.segmentCount):
        row = [i+1, results.segment(i).data[0], results.segment(i).data[1], results.segmentColor(i)]
        for var in vars:
            row.append( results[var][i] )
        tbl.append( row )
    return tbl

def ReportSegments(visVarsOnly):
    def ReportSegs(results, context):
        rep = EnsureWordReport(context)
        rep.writeTable( BuildSegmentsTable(results, visVarsOnly) )
    return ReportSegs

Reports.Add('Segments (all)', ReportSegments(False), ['MIL', 'MPL', 'SKM', 'AMP', 'HalfAmp', 'Stat', 'ChopIdl', 'IdlBase'])
Reports.Add('Segments (visible)', ReportSegments(True), ['MIL', 'MPL', 'SKM', 'AMP', 'HalfAmp', 'Stat', 'ChopIdl', 'IdlBase'])


def ReportColorMeans(visVarsOnly):
    def ReportColors(results, context):
        rep = EnsureWordReport(context)
        tbl = []
        colors = results.usedColors
        hdr = ['']
        for color in colors:
            hdr.append('%s mean'%str(color))
            hdr.append('%s sd'%str(color))
        tbl.append(hdr)
        if visVarsOnly:
            vars = results.visibleVars
        else:
            vars = results.keys()
        for var in vars:
            row = [var]
            for color in colors:
                row.append(results[var].mean(color))
                row.append(results[var].sd(color))
            tbl.append(row)
        rep.writeTable(tbl)

    return ReportColors        

Reports.Add('Color Means (all)', ReportColorMeans(False), ['MIL', 'MPL', 'MIP', 'Mac', 'SKM', 'AMP', 'HalfAmp', 'Stat', 'ChopIdl', 'IdlBase', 'MdlRpt'])
Reports.Add('Color Means (visible)', ReportColorMeans(True), ['MIL', 'MPL', 'MIP', 'Mac', 'SKM', 'AMP', 'HalfAmp', 'Stat', 'ChopIdl', 'IdlBase', 'MdlRpt'])


def MSearchTopN(result, n):
    top = []
    trialsPerModel = result['Properties']['TrialsPerModel'].data[0]
    i = 0 # unique seq number to stop comparison if aic equal
    trial = result.find('ModelTrial')
    while not trial.isNull:
        mtrials = []
        for mt in xrange(trialsPerModel):
            aic = trial.find(' AIC')
            if aic.isNull:
                aic = trial[' LL']
            mtrials.append( (aic.data[0], i, trial) )
            trial = trial.next('ModelTrial')
            i += 1
        mtrials.sort()
        heapq.heappush(top, mtrials[trialsPerModel-1])
        if len(top) > n:
            heapq.heappop(top)
    top.reverse()
    return top


def ReportMsearchTopTen(results, context):
    rep = EnsureWordReport(context)
    for aic, ii, trial in MSearchTopN(results.node, 10):
        mdl = trial['ModelFile'].next('ModelFile')
        rep.pasteModelImage( mdl,
                             Width=400, Height=100, Bitmap=0,
                             DrawFrame=1, ShowTitles=0, Landscape=1 )
        writeModelConstraints(rep, mdl)
        rep.write("LL: %.5f  Grad: %s" % (trial[" LL"].data[0], trial["Gradient"].data))
        errcode = trial["ErrorCode"].data[0]
        rep.write("Iterations: %d  Warnings: %d%s" % (trial["Iterations"].data[0], trial["Warnings"].data[0],
                                                      (errcode and ("  [ErrorCode: %d]" % errcode) or "")))
        rep.write()

Reports.Add('Top 10 Models', ReportMsearchTopTen, ['MSearch', 'Prtb'])
